# ## Explainable Conditional LSTM random walk graph GAN training
from eggen import utils
from eggen.eggen_shadow import *

import tensorflow as tf
import scipy.sparse as sp
import numpy as np
from matplotlib import pyplot as plt
from sklearn.metrics import roc_auc_score, average_precision_score
import time
import pandas as pd # data processing, CSV file I/O (e.g. pd.read_csv)
import re
import pickle

# ## Load data
## Load data
# """ Entire set of email data """
data = pd.read_csv('./data/Enron_messages.csv')
print("Data chunk size: " ,len(data))

#### Dept data (conditions)
cond_list = pd.read_csv('./data/Enron_DeptList.csv') #, sep=" ", header=None)
cond_list.columns = ['ID', 'DEPT']
cond_list = cond_list.values


import networkx as nx
import nxviz as nv

G = nx.from_pandas_edgelist(data, 'Sender ID', 'Receiver ID') #, edge_attr=['SENDER DE


## Preparing data
Adjtraining = nx.adjacency_matrix(G)
Adjtraining = sp.csr_matrix(Adjtraining, dtype='float64')
_A_obs = Adjtraining
_A_obs = _A_obs + _A_obs.T # (597, 597)
_A_obs[_A_obs > 1] = 1 # Max value of 1 (597, 597)

""" Reduce input graph to a subgraph where only the nodes in largest n_components are kept. """ 
lcc = utils.largest_connected_components(_A_obs) # len(lcc) = 584
_A_obs = _A_obs[lcc,:][:,lcc] # (584, 584)
_N = _A_obs.shape[0] # 584

""" Get the subset of cond_list before reindexing from 0 """
ind_nonlcc = np.delete(np.arange(len(cond_list)), lcc)
lcc_condlist = np.delete(cond_list, ind_nonlcc, 0)
print("Length of cond_list", len(cond_list))

# print("**** N nodes {:}, E edges: {:}, K classes: {:} ****".format(_N,
#                                                                   np.count_nonzero(_A_obs.todense()),
#                                                                   len(np.unique(lcc_condlist[:,1]))
#                                                                   ))
print("**** N nodes {:}, E edges: {:}, K classes: {:} ****".format(_N,
                                                                  np.count_nonzero(_A_obs.todense())/2,
                                                                  len(np.unique(lcc_condlist[:,1]))
                                                                  ))

#### Separate the edges into train, test, validation
val_share = 0.1
test_share = 0.05
seed = 2020 #481516234  
"""
Split the edges of the adjacency matrix into train, validation and test edges and randomly samples equal amount of validation and test non-edges. 
"""
train_ones, val_ones, val_zeros, test_ones, test_zeros = utils.train_val_test_split_adjacency(_A_obs, val_share, test_share, seed, undirected=True, connected=True, asserts=False) 

## EGGen
train_graph = sp.coo_matrix((np.ones(len(train_ones)),(train_ones[:,0], train_ones[:,1]))).tocsr()
assert (train_graph.toarray() == train_graph.toarray().T).all()


#### Parameters
""" Adjustable parameters for training. """ 
# setting GPU id 
gpu_id = 1
# setting the number of nodes
_N = _A_obs.shape[0]
# setting the length of random walks
rw_len = 16 #8 #32 #
# setting the training data batch size
batch_size = 128 #512 #
# getting the number of departments
n_conds=len(np.unique(lcc_condlist[:,1]))
print("n_conds: ", n_conds)
sample_batch = 128 #2048 #256 #512 #1024 #
log_num = 31 #99 #



walker = utils.RandomWalker(train_graph, lcc_condlist, rw_len, p=1, q=1, batch_size=batch_size, sample_batch=sample_batch)



#### ====== Create our generative model ======
l2_gen=1e-7; l2_disc=5e-5 #1e-4 
gencond_lay=[10]; gen_lay=[50]; disc_lay=[40] 
# lr_gencond=0.01; lr_gen=0.0003; lr_disc=0.0003 #0.0002
lr_gencond=0.01; lr_gen=0.0002; lr_disc=0.0002 #0.0002
gencond_iters=1; gen_iters=1; disc_iters=3
discWdown_size=128; genWdown_size=128 


eggen = EGGen(_N,
rw_len,
walk_generator=walker,
n_conds=n_conds,
condgenerator_layers=gencond_lay,
generator_layers=gen_lay,
discriminator_layers=disc_lay,
W_down_discriminator_size=discWdown_size,
W_down_generator_size=genWdown_size,
batch_size=batch_size,
sample_batch=sample_batch,
condition_dim=n_conds,
gencond_iters=gencond_iters,
gen_iters=gen_iters,
disc_iters=disc_iters,
wasserstein_penalty=10, 
l2_penalty_generator=l2_gen,
l2_penalty_discriminator=l2_disc,
lr_gencond=lr_gencond,
lr_gen=lr_gen,
lr_disc=lr_disc,
noise_dim=16, #32, #
noise_type="Gaussian", #"Uniform", #
temp_start=5.0, #30.0, #10.0, #
min_temperature=0.5,
temperature_decay=1-5e-5,
seed=15, #seed, #
use_gumbel=True,
legacy_generator=False,
gpu_id=gpu_id,
plot_show=False
)

# In[ ]:


# #### Define the stopping criterion
stopping_criterion = "val"
assert stopping_criterion in ["val", "eo"], "Please set the desired stopping criterion."
if stopping_criterion == "val": # use val criterion for early stopping
    stopping = None
elif stopping_criterion == "eo":  #use eo criterion for early stopping
    stopping = 0.5 # set the target edge overlap here


# #### Train the model
eval_every = plot_every = 1000 #200
max_iters = 20000 #10000 #
patience= 20 

# train and save model to ./snapshots/
log_dict = eggen.train(A_orig=_A_obs, val_ones=val_ones, val_zeros=val_zeros, stopping=stopping,
                        eval_every=eval_every, plot_every=plot_every, max_patience=patience, max_iters=max_iters)


# #### Save the training log
## when changing the directory, remember to change directory in eggen.train() too
# save_directory = "./testing"
save_directory = "./snapshots_shadow" #"./snapshots_gencond"  #"./snapshots_gencond2" 
model_name = "shadowgen"
# log_num = 1

save_log = "{}/log{}_{}_maxiter{}_evalevery{}.pkl".format(save_directory, log_num, model_name, max_iters, eval_every)
f = open(save_log,"wb")
pickle.dump(log_dict,f)
f.close()




# # ## ====== generate graphs to evaluate performance ======
# sample_many, explain_conds = eggen.generate_discrete(10000, conds=True, rw_len=rw_len, reuse=True) 



# dmax,dmin,deg,lcc,wc,cc,tc,sc,law,gini,rel,assrt,coe,ncomp,intra,inter,cpl,eo = ([] for i in range(18))
# num_trials = 5 #1 #
# num_paths = 5000 #5000

# def compute_stats(samples):
#     rws = np.array(samples).reshape([-1, rw_len])
#     print("rws: ", rws.shape)
#     scores_matrix = utils.score_matrix_from_random_walks(rws, _N).tocsr()
    
#     A_select = sp.csr_matrix((np.ones(len(train_ones)), (train_ones[:,0], train_ones[:,1])))
#     A_select = train_graph
#     sampled_graph = utils.graph_from_scores(scores_matrix, A_select.sum())
#     EO = utils.edge_overlap(A_select.toarray(), sampled_graph)/A_select.sum()
    
#     stats = utils.compute_graph_statistics(sampled_graph) #, encoded)
#     return stats, EO

# for trials in range(num_trials):
#     print("trial: ", trials)
#     start_time = time.time()
  
#     samples = []
#     for _ in range(num_paths): #6000):
#         if (_+1) % 1000 == 0:
#             print(_+1)
#         samples.append(sample_many.eval({eggen.tau: 0.5}))

#     stats, EO = compute_stats(samples)
#     print("Trial %i: --- %s seconds ---" % (trials, time.time() - start_time))

#     dmax.append(stats['d_max'])
#     dmin.append(stats['d_min'])
#     deg.append(stats['d'])
#     lcc.append(stats['LCC'])
#     wc.append(stats['wedge_count'])
#     cc.append(stats['claw_count'])
#     tc.append(stats['triangle_count'])
#     sc.append(stats['square_count'])
#     law.append(stats['power_law_exp'])
#     gini.append(stats['gini'])
#     rel.append(stats['rel_edge_distr_entropy'])
#     assrt.append(stats['assortativity'])
#     coe.append(stats['clustering_coefficient'])
#     ncomp.append(stats['n_components'])
#     cpl.append(stats['cpl'])
#     eo.append(EO)
#     print("stats: ", stats)


# # all_stats = [dmax, dmin, deg, lcc, wc, cc, tc, sc, law, gini, rel, assrt, coe, ncomp, cpl, eo]
# # ====== ASST CLUST CPL GINI MD PLE EO ====== 
# all_stats = [assrt, coe, cpl, gini, dmax, law, eo, dmin, deg, lcc, wc, cc, tc, sc, rel, ncomp]

# # avg_stats = [np.mean(dmax), np.mean(dmin), np.mean(deg), np.mean(lcc), np.mean(wc), np.mean(cc), np.mean(tc), np.mean(sc), np.mean(law), np.mean(gini), np.mean(rel), np.mean(assrt), np.mean(coe), np.mean(ncomp), np.mean(cpl), np.mean(eo)]
# avg_stats = [np.mean(assrt), np.mean(coe), np.mean(cpl), np.mean(gini), np.mean(dmax), np.mean(law), np.mean(eo), np.mean(dmin), np.mean(deg), np.mean(lcc), np.mean(wc), np.mean(cc), np.mean(tc), np.mean(sc), np.mean(rel), np.mean(ncomp)]
# print("avg_stats: ", avg_stats)


# from scipy import stats
# # stderror_stats = [stats.sem(dmax), stats.sem(dmin), stats.sem(deg), stats.sem(lcc), stats.sem(wc), stats.sem(cc), stats.sem(tc), stats.sem(sc), stats.sem(law), stats.sem(gini), stats.sem(rel), stats.sem(assrt), stats.sem(coe), stats.sem(ncomp), stats.sem(cpl), stats.sem(eo)]
# stderror_stats = [stats.sem(assrt), stats.sem(coe), stats.sem(cpl), stats.sem(gini), stats.sem(dmax), stats.sem(law), stats.sem(eo), stats.sem(dmin), stats.sem(deg), stats.sem(lcc), stats.sem(wc), stats.sem(cc), stats.sem(tc), stats.sem(sc), stats.sem(rel), stats.sem(ncomp)]
    
# save_directory = "./generate_stats" #"./snapshots_gencond"  #"./snapshots_gencond2" 
# # log_num = 0
# data_name  = "EUcore-top" #"EUcore" #

# save_stats = "{}/{}_stats{}.txt".format(save_directory, data_name, log_num)

# np.savetxt(save_stats, np.c_[avg_stats]) #,stderror_stats, all_stats])



